# -*- coding: utf-8 -*-
# CPPDoc Plugin
# vim: ts=4 nowrap textwidth=80
# Copyright © 2010,2012 Peter Bakota
#

from gi.repository import GObject, Gedit, Gtk, Gdk
from datetime import datetime
import re,os

class CPPDocPlugin(GObject.Object, Gedit.WindowActivatable):
    __gtype_name = "CppDocPlugin"
    window = GObject.property(type=Gedit.Window)
    handler_ids = []

    def __init__(self):
        GObject.Object.__init__(self)

    def do_activate(self):
        view = self.window.get_active_view()
        if view:
            self.setup_doc(view, 'none')

    def do_deactivate(self):
        for handler_id, view in self.handler_ids:
            view.disconnect(handler_id)

    def do_update_state(self):
        view = self.window.get_active_view()
        lang = 'none'
        if view:
            buf = view.get_buffer()
            language = buf.get_language()
            if language:
                lang = language.get_id()
            self.setup_doc(view, lang)

    def setup_doc(self, view, lang):
        if getattr(view, 'cppdoc_instance', False) == False:
            setattr(view, 'cppdoc_instance', CppDoc())
            handler_id = view.connect('key-press-event', 
                view.cppdoc_instance.key_press_handler)
            self.handler_ids.append((handler_id, view))
        view.cppdoc_instance.set_language(lang)

class CppDoc:

    def __init__(self):
        self.__not_available = True
        self.re_trigger = None
        self.re_comment = None
        self.re_var = None
        self.re_class = None
        self.re_function = None
        self.re_module_start = None
        return

    def set_language(self, lang):
        #print "Lang: ",lang
        self.__not_available = False
        if lang == 'none':
            self.__not_available = True
        elif lang == 'cpp' or lang == 'chdr':
            self.preceding_white_space_pattern = re.compile(r'^(\s*)')
            self.re_trigger = re.compile(r'^\s*/\*\*$'); # CppDoc trigger
            self.re_comment = re.compile(r'^\s*\*\s*(?!\/)'); # Cpp comment trigger
            self.re_function = re.compile(r'\s*(?P<SCOPE>inline|static)*\s*(?P<RETURN>[0-9a-zA-Z_\*&:]+)*?\s*(?P<FUNCNAME>[^\s]+)\s*\(\s*(?P<PARAMS>[^\)]+)?\s*\)\s*[;\{]*')
            self.re_var = re.compile(r'\s*(?P<SCOPE>static)*\s*(?P<TYPE>[0-9a-zA-Z\*:\<\>\,]+)\s*(?P<VARNAME>[0-9a-zA-Z:_\*]+)\s*=*\s*(?P<VALUE>[^;]+)?;')
        else:
            self.__not_available = True

    def __get_current_line(self, view, buf):
        cursor_iter = buf.get_iter_at_mark(buf.get_insert())
        line_start_iter = cursor_iter.copy()
        view.backward_display_line_start(line_start_iter)
        return buf.get_text(line_start_iter, cursor_iter, True)

    def __get_bellow_line(self, view, buf):
        line_start_iter = buf.get_iter_at_mark(buf.get_insert())
        line_start_iter.forward_line()
        line_start_iter.set_line_offset(0)
        line_end_iter = line_start_iter.copy()
        line_end_iter.forward_line()
        line_end_iter.backward_char()
        return buf.get_text(line_start_iter, line_end_iter, True)

    def __get_above_line(self, view, buf):
        line_start_iter = buf.get_iter_at_mark(buf.get_insert())
        line_start_iter.backward_line()
        line_start_iter.set_line_offset(0)
        line_end_iter = buf.get_iter_at_mark(buf.get_insert())
        line_end_iter.set_line_offset(0)
        line_end_iter.backward_char()
        return buf.get_text(line_start_iter, line_end_iter, True)

    def key_press_handler(self, view, event):
        if (self.__not_available) or (event.keyval != Gdk.KEY_Return): # Key trigger
            return
        buf = view.get_buffer()
        s = self.__get_current_line(view, buf)
        if self.re_trigger.match(s):
            bellow = self.__get_bellow_line(view, buf)
            preceding_white_space = self.preceding_white_space_pattern.search(bellow).groups()[0]
            m = self.re_function.match(bellow)
            if m:
                #print "Function: ", m.groups()
                buf.insert_at_cursor('\n')
                buf.insert_at_cursor(preceding_white_space)
                buf.insert_at_cursor(' * %s\n' % m.group('FUNCNAME'))
                buf.insert_at_cursor(preceding_white_space)
                buf.insert_at_cursor(' *\n')
                if m.group("SCOPE") is not None:
                    buf.insert_at_cursor(preceding_white_space)
                    buf.insert_at_cursor(' * @access %s\n' % m.group("SCOPE"))
                if m.group("PARAMS") is not None:
                    for param in m.group("PARAMS").split(","):
                        param = param.strip()
                        if param == '':
                            continue
                        try:
                            i = param.rfind(" ")
                            type = param[0:i]
                            name = param[i+1:]
                        except ValueError:
                            type = "unknown"
                            name = param
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @param %s %s\n' % (type,name))
                if m.group('FUNCNAME')[:2]!="::" and m.group('RETURN') is not None:
                    buf.insert_at_cursor(preceding_white_space)
                    buf.insert_at_cursor(' * @return %s\n' % m.group('RETURN'))
                    buf.insert_at_cursor(preceding_white_space)
                    buf.insert_at_cursor(' *\n')
                buf.insert_at_cursor(preceding_white_space)
                buf.insert_at_cursor(' * @brief \n')
                buf.insert_at_cursor(preceding_white_space)
                buf.insert_at_cursor(' */')
            else:
                m = self.re_var.match(bellow)
                if m is not None:
                    #print "Var: ", m.groups()
                    buf.insert_at_cursor('\n')
                    if m.group("SCOPE") is not None:
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @access %s\n' % m.group("SCOPE"))
                    buf.insert_at_cursor(preceding_white_space)
                    buf.insert_at_cursor(' * @brief %s %s\n' % (m.groups("TYPE"), m.group("VARNAME")))
                    buf.insert_at_cursor(preceding_white_space)
                    buf.insert_at_cursor(' */')
                else:
                    lineno = buf.get_iter_at_mark(buf.get_insert()).get_line()
                    #print "lineno is: [%d]" % lineno
                    if lineno==0:
                        buf.insert_at_cursor('\n')
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @file %s\n' % buf.get_short_name_for_display())
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' *\n')
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @author %s\n' % os.environ['USER'])
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @date %s\n' % datetime.today().strftime('%c')) # see: http://docs.python.org/library/datetime.html
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' * @brief \n')
                        buf.insert_at_cursor(preceding_white_space)
                        buf.insert_at_cursor(' */')
                    else:
                        return
            return True
        elif self.re_comment.match(s):
            # Continue in comment mode
            buf.insert_at_cursor('\n' + s[:s.find('*')] + '* ')
            return 

